@@ -5,12 +5,13 @@
 Utilities for sending files over ssh using the scp1 protocol.
 """
 
-__version__ = '0.11.0'
+__version__ = '0.13.0'
 
 import locale
 import os
 import re
 from socket import timeout as SocketTimeout
+import types
 
 
 # this is quote from the shlex module, added in py3.3
@@ -84,7 +85,7 @@ class SCPClient(object):
     (matching scp behaviour), but we make no attempt at symlinked directories.
     """
     def __init__(self, transport, buff_size=16384, socket_timeout=10.0,
-                 progress=None, sanitize=_sh_quote):
+                 progress=None, progress4=None, sanitize=_sh_quote):
         """
         Create an scp1 client.
 
@@ -96,21 +97,31 @@ class SCPClient(object):
         @type socket_timeout: float
         @param progress: callback - called with (filename, size, sent) during
             transfers
+        @param progress4: callback - called with (filename, size, sent, peername)
+            during transfers. peername is a tuple contains (IP, PORT)
         @param sanitize: function - called with filename, should return
             safe or escaped string.  Uses _sh_quote by default.
-        @type progress: function(string, int, int)
+        @type progress: function(string, int, int, tuple)
         """
         self.transport = transport
         self.buff_size = buff_size
         self.socket_timeout = socket_timeout
         self.channel = None
         self.preserve_times = False
-        self._progress = progress
+        if progress is not None and progress4 is not None:
+            raise TypeError("You may only set one of progress, progress4")
+        elif progress4 is not None:
+            self._progress = progress4
+        elif progress is not None:
+            self._progress = lambda *a: progress(*a[:3])
+        else:
+            self._progress = None
         self._recv_dir = b''
         self._rename = False
         self._utime = None
         self.sanitize = sanitize
         self._dirtimes = {}
+        self.peername = self.transport.getpeername()
 
     def __enter__(self):
         self.channel = self._open()
@@ -155,7 +166,7 @@ class SCPClient(object):
 
         self.close()
 
-    def putfo(self, fl, remote_path, mode='0644'):
+    def putfo(self, fl, remote_path, mode='0644', size=None):
         """
         Transfer file-like object to remote host.
 
@@ -165,13 +176,21 @@ class SCPClient(object):
         @type remote_path: str
         @param mode: permissions (posix-style) for the uploaded file
         @type mode: str
+        @param size: size of the file in bytes. If ``None``, the size will be
+            computed using `seek()` and `tell()`.
         """
+        if size is None:
+            pos = fl.tell()
+            fl.seek(0, os.SEEK_END)  # Seek to end
+            size = fl.tell() - pos
+            fl.seek(pos, os.SEEK_SET)  # Seek back
+
         self.channel = self._open()
         self.channel.settimeout(self.socket_timeout)
         self.channel.exec_command(b'scp -t ' +
                                   self.sanitize(asbytes(remote_path)))
         self._recv_confirm()
-        self._send_file(fl, remote_path, mode, size=len(fl.getvalue()))
+        self._send_file(fl, remote_path, mode, size=size)
         self.close()
 
     def get(self, remote_path, local_path='',
@@ -262,16 +281,16 @@ class SCPClient(object):
         if self._progress:
             if size == 0:
                 # avoid divide-by-zero
-                self._progress(basename, 1, 1)
+                self._progress(basename, 1, 1, self.peername)
             else:
-                self._progress(basename, size, 0)
+                self._progress(basename, size, 0, self.peername)
         buff_size = self.buff_size
         chan = self.channel
         while file_pos < size:
             chan.sendall(fl.read(buff_size))
             file_pos = fl.tell()
             if self._progress:
-                self._progress(basename, size, file_pos)
+                self._progress(basename, size, file_pos, self.peername)
         chan.sendall('\x00')
         self._recv_confirm()
 
@@ -411,9 +430,9 @@ class SCPClient(object):
         if self._progress:
             if size == 0:
                 # avoid divide-by-zero
-                self._progress(path, 1, 1)
+                self._progress(path, 1, 1, self.peername)
             else:
-                self._progress(path, size, 0)
+                self._progress(path, size, 0, self.peername)
         buff_size = self.buff_size
         pos = 0
         chan.send(b'\x00')
@@ -425,8 +444,7 @@ class SCPClient(object):
                 file_hdl.write(chan.recv(buff_size))
                 pos = file_hdl.tell()
                 if self._progress:
-                    self._progress(path, size, pos)
-
+                    self._progress(path, size, pos, self.peername)
             msg = chan.recv(512)
             if msg and msg[0:1] != b'\x00':
                 raise SCPException(asunicode(msg[1:]))
